/**
 * @file Fonctions pour interagir avec la BDD sigma et le LDAP.
 * @author akka vodol
 */
import knex from '../../../db/knex_router';
import { renseignerSurUtilisateur, repliquerTOLdesIds, 
    listerGroupes, listerMembres, listerAdministrateurs 
} from '../../ldap/ldap_data';
import { exportAllDeclaration } from 'babel-types';

export { renseignerSurUtilisateur, repliquerTOLdesIds, listerMembres };

/*
Le tag @rights et la gestion des authorisations

Le système GraphQL est pensé comme l'interface par laquelle les utilisateurs 
intéragissent avec sigma, les graphismes en moins.
Pour cette raison, et pour des questions de sécurité, il faut partir du principe que
l'utilisateir peut rédiger n'importe quelle requête, et que c'est au niveau des resolvers
que la sécurité ce déroule. C'est dans cette optique que j'utilise le tag @rights

Commençons par un rappel sur le fonctionnement des droits. 
Chaque utilisateur a un certain niveau de droit sur chaque groupe. Ce niveau de droit indique
ce qu'il a le droit de savoir et de faire. Chaque niveau est inclu dans les niveaus supérieur.
Les différends niveaux sont :
none - aucun droit
viewer : l'utilisateur a visibilité sur le groupe. Il sait que le groupe existe, et a accès à un certain nombre d'infos.
member : l'utilisateur est membre du groupe
speaker : l'utilisateur peut parler au nom du groupe. Il a le droit de publier des annonces et d'organiser des évènements
admin : l'utilisateur a tous les droits sur le groupe

Certaines fonctions de connectors effectuent des vérifications d'authorisations avant 
de renvoyer une réponse, d'autres non. Pour être sur qu'on ne renvoie jamais de réponse
sans avoir au préalable éffectué les bonnes vérifications, chaque fonction possède dans sa
description un attribut droit, qui décrit les droits que fournit cette fonction.

La valeur de @rights peut être :
super - la fonction n'effectue aucune vérification, et renvoie le resultat demandé
admin( groupUID ) - la fonction ne fait que ce qu'un admin du groupe indiqué aurait le droit de faire
speaker( groupUID ), member( groupUID ), veiwer( groupUID ) - même chose
user - la fonction ne fait que ce que l'utiliateur a le droit de faire (vérifications via l'argument user)

La procédure a suivre est la suivante : quand une fonction possède un certain niveau de droit, 
elle ne peut appeler une fonction possédant un niveau de droit plus large que si 
1 ) on a au préalable vérifié que l'utilisateur possédait effectivement ces droits. 
ou
2 ) on s'est assuré que l'opération effectuée par cet appel particulier de la fonction était dans les droits
de l'utilisateur

Les resolvers de base de mutation et query ont des droits user.

Les fonctions qui ne modifient pas la BDD et ne renvoient pas de données sur la BDD n'ont pas de rights.
*/




/**
 * @summary Génère une promise.
 * @function
 * @desc Les fonctions ici sont toutes supposées renvoyer une promise. 
 * Si on veut renvoyer une valeur directement, cette fonction permet de construire 
 * une Promise qui renvoie cette valeur facilement.
 * @arg {Object} val - La valeur qu'on veut renvoyer.
 * @return {Promise(Object)} Une promise qui renvoi val
 */
const quickPromise = (val) => {
    return new Promise( (resolve, reject) => {
        resolve(val);
    });
};

/**
 * @summary Renvoie le type d'un groupe.
 * @function
 * @desc Parcours les BDD pour savoir dans laquelle se trouve le groupe ayant l'UID donné. 
 * Cette opération nécéssite un parcours de base de donnée, et il est préférable de ne pas 
 * sans servir si on a un autre moyend de connaitre le typed d'un groupe
 * (Par exemple, si on dispose d'un retour de BDD pour ce groupe. Le champ 'type' indique alors son type.)
 * @arg {Object} user - Objet contenant un attribut `uid` de type `string`. 
 * User représente l'utilisateur qui a effectué la requête. 
 * @arg {Object} groupUID - L'id du groupe dont on veut connaître le type.
 * @return {Promise(String)} Un string représentant le type du groupe.
 * Peut être "SimpleGroup" ou "MetaGroup". Renvoie `Undefined` si le groupe n'existe pas
 * @rights super
 */
export const getGroupType = (user, groupUID) => {
    return knex('simple_groups').select('uid').where('uid', groupUID).then( sg_res => {
        if(sg_res)
            return "SimpleGroup";
        return knex('meta_groups').select('uid').where('uid', groupUID).then(mg_res => {
            if(mg_res)
                return "MetaGroup";
            return undefined;
        });
    });
};

/**
 * @summary Renvoie tous les utilisateurs ayant des droits d'administrateur sur un groupe.
 * @function
 * @desc Les utilisateurs qui ont un droit d'administrateur sur un groupe simple sont ses administrateurs 
 * et les utilisateurs ayant droit d'admin sur son parent.
 * Les utilisateurs ayant droit d'admin sur un meta-groupe sont les utilisateurs 
 * ayant droit d'admin sur un des groupes membres
 * @arg {String} uid - L'uid du groupe dont on veut les administrateurs. 
 * @return {Promise} Retour de requête knex. Promise qui renvera une liste 
 * de tous les utilisateurs ayant droit d'admin sur le groupe
 * rights member(groupUID)
 */
export const getUsersWithAdminRights = (user, groupUID) => {
    return getGroupType(user, groupUID).then( groupType => {
        
        if(groupType == "SimpleGroup"){
            return listerAdministrateurs(user, groupUID).then(adminList => {
                if (typeof adminList == "undefined")
                    return undefined;
                else
                    return knex('simple_groups').select('parentuid').where('uid', groupUID).then(req => {
                        if (req[0].parentuid)
                            return getUsersWithAdminRights(user, req[0].parentuid).then(parentAdmins => {
                                return adminList.concat(parentAdmins);
                            });
                        else
                            // pour les besoins des tests, anatole romon a tout les droits
                            return adminList.concat(['anatole.romon']); 
                    });
            });
        }else{
            return getMetaGroupAdminMembers(user, groupUID).then(mg_members => {
                function fuse (l){
                    if(l)
                        return getUsersWithAdminRights(l.pop()).then(usrs => fuse(l).concat(usrs) );
                    else
                        return [];
                }
            });
        }
    });
};

/**
 * @summary teste si un utilisateur a des droits
 * @desc Cette fonction effectue une requête knex. Elle gère l'arête de parenté.
 * @arg {Object} user - Objet contenant un attribut `uid` de type `string`. 
 * User représente l'utilisateur qui a effectué la requête. 
 * @return {Promise} Retour de requête knex. Liste de tous les groupes que l'utilisateur a le droit de voire.
 * rights user
 */
export const hasAdminRights = (user, groupUID) => {
    if(user.uid == "anatole.romon")
        return quickPromise(true);
    return getUsersWithAdminRights(user, groupUID).then(adminList => {
        return (typeof adminList != "undefined" && adminList.indexOf(user.uid) != -1);
    });
};


/**
 * @summary Renvoie une liste des id de tous les groupes visibles par l'utilisateur
 * @desc Cette fonction génère un callback qui créé une table contenant les uid de tous les groupes visibles
 * @arg {Object} user - Objet contenant un attribut `uid` de type `string`. 
 * User représente l'utilisateur qui a effectué la requête. 
 * @return {Promise(Callback)} callback contruisant une requête knex pour une table de tous les id visibles.
 * @rights user
 */
export const getVisibleGroupCallback = (user) => {
    return listerGroupes(user, user.uid).then(group_ids => {
        return function (global_query_builder){
            if (typeof group_ids == "undefined")
                throw "invalid user";
            var membered_groups = qb => qb.select('simple_groups.uid').from('simple_groups').whereIn('uid', group_ids.concat(['kes']));
            var directly_visible_simple_groups = qb =>  qb.with('membered_groups', membered_groups).select('simple_groups.uid').from('simple_groups').distinct()
                .innerJoin('membered_groups',
                    function () {
                        this.on('simple_groups.uid', '=', 'membered_groups.uid')
                            .orOn('simple_groups.parentuid', '=', 'membered_groups.uid');
                    }
                );
            return directly_visible_simple_groups(global_query_builder);
        };
    });
    
};

/**
 * @summary Renvoie le nom de la table dans laquelle il faut chercher en fonction de ce qu'on veut
 * @desc a remplir
 * @arg {String} wantedType - Un string indiquant le type de groupe qu'on veut. Peut être `"simple"`, `"meta"` ou `"all"`. 
 * @return {String} Le nom de la table dans laquelle la requète doit être effectuée.
 */
function getGroupTableName(wantedType){
    switch(wantedType){
    case "simple":
        return "simple_groups";
    case "SimpleGroup":
        return "simple_groups";
    case "meta":
        return "meta_groups";
    case "MetaGroup":
        return "meta_groups";
    case "all":
        return "groups";
    default:
        throw "invalid type request : " + wantedType + "is not a valid group type";
    }
}

/**
 * @summary Renvoie un unique groupe, ssi ce groupe est visible par l'utilisateur
 * @desc Pour l'instant, la fonction effectue la même requête que `getAllVisibleGroups` 
 * et restreint au groupe demandé. Cette fonction peut être implémentée de manière 
 * plus efficace et plus chiante.
 * @arg {Object} user - Utilisateur effectuant la requête. 
 * @arg {String} uid - Identifiant du groupe voulu.
 * @arg {String} type - Type de groupe voulu. `"simple"`, `"meta"` ou `"all"`.
 * @return {Promise(group)} Retour de requête knex. Le groupe demandé, si l'utilisateur a le droit de la voire.
 * @rights user
 */
export async function getGroupIfVisible(user, groupUID, type="all"){
    let group_table_name = getGroupTableName(type);
    let visible_groups = await getVisibleGroupCallback(user);
    return knex.with('visible_groups', visible_groups).select()
        .from(group_table_name).innerJoin('visible_groups', function (){
            this.on('visible_groups.uid', '=', group_table_name + '.uid');
        }).where(group_table_name + '.uid', groupUID).then(res => res[0]);
}

export const getSimpleGroupIfVisible = (user, groupUID) => getGroupIfVisible(user, groupUID, "simple");
export const getMetaGroupIfVisible = (user, groupUID) => getGroupIfVisible(user, groupUID, "meta");

export async function getAllVisibleSimpleGroups (user){
    let visible_groups = await getVisibleGroupCallback(user);
    return knex.with('visible_groups', visible_groups).select().from("simple_groups").innerJoin('visible_groups', function (){
        this.on('visible_groups.uid', '=', 'simple_groups.uid');
    });
}

export async function getAllVisibleMetaGroups (user){
    let visible_groups = await getVisibleGroupCallback(user);
    return knex.with('visible_groups', visible_groups).select().from("meta_groups").innerJoin('visible_groups', function (){
        this.on('visible_groups.uid', '=', 'meta_groups.uid');
    });
}

/**
 * @summary Renvoie tous les groupes visibles par l'utilisateur user
 * @desc Cette fonction effectue une requête knex. Elle gère l'arête de parenté.
 * @arg {Object} user - Représente l'utilisateur qui a effectué la requête. 
 * @arg {String} wantedType - Type de groupe voulu : `"simple"`, `"meta"` ou `"all"`. 
 * @return {Promise} Retour de requête knex. Liste de tous les groupes que l'utilisateur a le droit de voire.
 * @rights user
 */
export async function getAllVisibleGroups(user){
    let all_simple_groups = await getAllVisibleSimpleGroups(user);
    let all_meta_groups = await getAllVisibleMetaGroups(user);
    return all_simple_groups.concat(all_meta_groups);
}

/**
 * @summary Teste si un utilisateur est membre d'un groupe
 * @arg {Object} user - Représente l'utilisateur qui a effectué la requête. 
 * @arg {Object} groupUID - L'id du groupe dont on veu savoir si l'utilisateur est membre. 
 * @return {Promise(Boolean)} Boolean indiquant si l'utilisateur est membre du groupe.
 * @rights user
 */
export async function isMember(user, groupUID){
    let member_list = await getGroupMemberUsers(user, groupUID);
    return member_list && member_list.indexOf(groupUID) != -1;
};

/**
 * @summary Attribue un UID qui n'a pas encore été utilisé à un groupe
 * @desc RASifie le string initialUID si necessaire (ramené à de l'ASCCI sans espace), puis si l'uid est deja pris rajoute un n a la fin et reteste
 * @arg {String} uid - L'uid du groupe dont on veut les administrateurs. 
 * @return {Promise} Retour de requête knex. Promise qui renvera une liste de tous les utilisateurs ayant droit d'admin sur le groupe
 * @rights user
 * remarque : n'importe qui peut tester si un groupe existe en demandant a créer un groupe avec ce nom la et en regardant si
 * son UID a été modifié. Je ne vois pas comment contourner ce problème, c'est donc une faille permanente de sigma.
 */
export function getAvailablegroupUID(initialUID){
    let rasUID = initialUID.replace(' ', '_').replace(/\W/g, '').toLowerCase();
    return knex.from('groups').where('uid', rasUID).then(res => {
        if (res.length == 0) {
            return (rasUID);
        } else {
            return (getAvailablegroupUID(rasUID + 'n'));
        }
    });
}

/**
 * @summary Créé un groupe si les arguments sont tous valides
 * @desc Les arguments doivent être valides, sauf pour uid. Une clé uid valide sera générée dans tous les cas. 
 * Les authorisations de l'utilisateur ne sont pas vérifiées
 * On teste si l'utilisateur qui envoie la requête a des droits d'admin sur le parent du groupe qui doit être créé, avec la fonction
 * `getUsersWithAdminRights`.
 * Si un argument est invalide ou si l'utilisateur n'a pas les droits, la fonction renvoie une erreur
 * @arg {Object} user - L'utilisateur qui effectue la requête. 
 * @arg {Object} args - Les arguments envoyés à la mutation. Cf le schéma GraphQL 
 * @return {Promise} Retour de requête knex. Le groupe qui vient d'être créé. En cas d'echec, renvoie une erreur.
 * @rights admin (args.parentuid)
 */
export async function createSubgroup(user, args){
    if (typeof args.parentuid != 'string')
        throw "Illegal argument : parentuid must be a non null string";
    if (typeof args.name != 'string')
        throw "Illegal argument : name must be a non null string";

    let rasUID = await getAvailablegroupUID(args.uid);

    // TODO : appeller une fonction de ldap_data pour y créer un groupe.
    await knex('simple_groups').insert({
        uid: rasUID,
        parentuid: args.parentuid,
        createdAt: knex.fn.now(),
        updatedAt: this.createdAt,
        name: args.name,
        website: args.website,
        description: args.description,
        school: args.school,
        type : "simple"
    });

    return getGroupIfVisible(user, rasUID);
}

/**
 * @summary Créé un groupe si les arguments sont tous valides et l'utilisateur est authorisé
 * @desc Les arguments doivent être valides, sauf pour uid. Une clé uid valide sera générée dans tous les cas. 
 * On teste si l'utilisateur qui envoie la requête a des droits d'admin sur le parent du groupe qui doit être créé, avec la fonction `getUsersWithAdminRights`
 * Si un argument est invalide ou si l'utilisateur n'a pas les droits, la fonction renvoie une erreur
 * @arg {Object} user - L'utilisateur qui effectue la requête. 
 * @arg {Object} args - Les arguments envoyés à la mutation. Cf le schéma GraphQL 
 * @return {Promise} Retour de requête knex. Le groupe qui vient d'être créé. En cas d'echec, renvoie une erreur.
 * @rights user
 */
export async function createGroupIfLegal(user, args){
    if( await hasAdminRights(user, args.parentuid) ){
        return createSubgroup(user, args);
    }else{
        throw "illegal request : you must have admin rights over a group to create a subgroup of that group";
    }
}

/**
 * @summary Renvoie toues les requêtes de type UserJoinGroup 
 * @desc Une requête UserJoinGroup est envoyée par un utilisateur à un groupe, 
 * pour demander à rejoindre ce groupe
 * @arg {Object} user - L'utilisateur qui effectue la requête. 
 * @arg {String} args - L'identifiant du groupe qui reçoit la requête. 
 * @return {Promise(Object)} Retour de requête knex. Toutes les requêtes destinées au groupe.
 * @rights admin(recipientUID)
 */
export function getUserJoinGroupRequests(user, recipientUID){
    return knex.select('id', 'useruid', 'message').from('user_join_group')
        .where('recipient', recipientUID);
}

/**
 * @summary Renvoie toues les requêtes de type GroupJoinEvent 
 * @desc Une requête UserJoinGroup est envoyée par un groupe à un évènement (donc aux administrateurs de l'évènement), 
 * pour demander à rejoindre cet évènement.
 * Remarque : toutes les requêtes ont pour le moment un attribut recipient, 
 * mais ici il ne sera a terme pas utilisé.
 * @arg {Object} user - L'utilisateur qui effectue la requête. 
 * @arg {String} args - L'identifiant du groupe qui reçoit la requête. 
 * @return {Promise(Object)} Retour de requête knex. Toutes les requêtes destinées au groupe.
 * @rights speaker(recipientUID)
 */
export function getGroupJoinEventRequests(user, recipientUID){
    return knex.select('id', 'senderuid', 'eventuid', 'message').from('group_join_event')
        .where('recipient', recipientUID);
}


/**
 * @summary Renvoie toues les requêtes de type GroupJoinEvent 
 * @desc Une requête UserJoinGroup est envoyée par un groupe à un évènement (donc aux administrateurs de l'évènement), 
 * pour demander à rejoindre cet évènement.
 * Remarque : toutes les requêtes ont pour le moment un attribut recipient, 
 * mais ici il ne sera a terme pas utilisé.
 * @arg {Object} user - L'utilisateur qui effectue la requête. 
 * @arg {String} args - L'identifiant du groupe qui reçoit la requête. 
 * @return {Promise(Object)} Retour de requête knex. Toutes les requêtes destinées au groupe.
 * @rights speaker(recipientUID)
 */
export const getYourGroupHostEventRequests = (user, recipientUID) => {
    return knex.select('id', 'senderuid', 'eventuid', 'message').from('your_group_host_event')
        .where('recipient', recipientUID);
};


//Don't forget the argument user is the guy who makes the request, not the user we want
export const getUser = (user, userUID) => { 
    const refactorer = (data) => {
        console.log("type of nick",typeof data.nickname);
        if (typeof data.brRoom == 'string') data.brRoom = [data.brRoom];

        return {
            uid: userUID,
            lastName: data.sn,
            givenName: data.givenName,
            nickname: data.displayName,
            nationality: data.country,
            birthdate: data.brBirthdate,
            groups: data.brMemberOf,
            mail: data.mail,
            phone: data.telephoneNumber,
            address: data.brRoom,
            promotion: data.brPromo
        };
    };

    const result = renseignerSurUtilisateur(user, userUID).then(res => {
        console.log(res[0]);
        return refactorer(res[0]);
    });

    return result;
};

/**
 * @function Renvoie un événement en fonction de son identifiant.
 * @param {*} user - Utilisateur effectuant la requête.
 * @param {*} eventID - Identifiant unique de l'événement.
 */
export const getEvent = (user, eventID) => {
    return quickPromise(null);
};

/**
 * @summary Renvoie simplement un groupe en fonction de son identifiant.
 * @param {Object} user - Utilisateur effectuant la requête.
 * @param {String} groupUID - Identifiant unique du groupe.
 * @author manifold 
 * @rights super
 */
export const getGroup = (user, groupUID) => {
    // Une sélection sur une table renvoie un tableau.
    // Knex renvoie une promesse, qui se résout en le tableau sélectionné.
    // On récupère son unique valeur, puisqu'on filtre sur l'identifiant unique.
    return knex.select().from('groups').where('uid',groupUID).then(results => results [0]);
};

/**
 * @summary Renvoie simplement un groupe simple en fonction de son identifiant.
 * @param {Object} user - Utilisateur effectuant la requête.
 * @param {String} groupUID - Identifiant unique du groupe.
 * @author manifold 
 * @rights super
 */
export const getSimpleGroup = (user, groupUID) => {
    return knex.select().from('simple_groups').where('uid',groupUID).then(results => results [0]);
};

/**
 * @summary Renvoie simplement un meta groupe en fonction de son identifiant.
 * @param {Object} user - Utilisateur effectuant la requête.
 * @param {String} groupUID - Identifiant unique du groupe.
 * @author manifold 
 * @rights super
 */
export const getMetaGroup = (user, groupUID) => {
    return knex.select().from('meta_groups').where('uid',groupUID).then(results => results [0]);
};

export const getMetaGroupAdminMembers = (user, metaGroupUID) => {
    return quickPromise([]);
};

/**
 * @summary Renvoie un callback qui génère tous les membres d'un meta-groupe.
 * @param {Object} user - Utilisateur effectuant la requête.
 * @param {String} metaGroupUID - Identifiant unique du groupe.
 * @return {Promise(callback)} a callback to build a query for the members of a group
 * It doesn't need to be a promise, but I figure having all of my functions return promises is 
 * easier than keeping track of which functions do and do not return promises.
 * @author akka vodol 
 * @rights member(metaGroupUID)
 */
export async function getMetaGroupMembersCallback(user, metaGroupUID){
    return function(query_builder){
        return query_builder.distinct().select().from('groups')
            .innerJoin('meta_group_membership', 'groups.uid', 'meta_group_membership.member_uid')
            .where('meta_group_membership.union_uid', '=', metaGroupUID);
    };
};

/**
 * @summary Renvoie un callback qui génère tous les membres admin d'un meta-groupe.
 * @param {Object} user - Utilisateur effectuant la requête.
 * @param {String} metaGroupUID - Identifiant unique du groupe.
 * @return {Promise(callback)} a callback to build a query for the admin members of a group
 * @author akka vodol 
 * @rights member(metaGroupUID)
 */
export async function getMetaGroupAdminMembersCallback(user, metaGroupUID){
    return function(query_builder){
        return query_builder.distinct().select().from('groups')
            .innerJoin('meta_group_membership', 'groups.uid', 'meta_group_membership.member_uid')
            .where('meta_group_membership.union_uid', '=', metaGroupUID)
            .where('meta_group_membership.status', '=', 'admin');
    };
};

/**
 * @summary Renvoie les membres d'un groupe quelquonque.
 * @param {Object} user - Utilisateur effectuant la requête.
 * @param {String} metaGroupUID - Identifiant unique du groupe.
 * @return {Promise(List)} Une liste des uid de tous les membres du groupe
 * @author akka vodol 
 * @rights member(metaGroupUID)
 */
export async function getGroupMemberUsers(user, GroupUID){
    switch( await getGroupType(user, GroupUID) ){
    case "simple":
        return listerMembres(user, GroupUID);
    case "meta":
        return getMetaGroupMemberUsers(user, GroupUID);
    default:
        return undefined;
    }
};

/**
 * @summary Renvoie les membres d'un meta groupe.
 * @param {Object} user - Utilisateur effectuant la requête.
 * @param {String} metaGroupUID - Identifiant unique du groupe.
 * @return {Promise(List)} Une liste des uid de tous les membres du groupe
 * @author akka vodol 
 * @rights member(metaGroupUID)
 */
export async function getMetaGroupMemberUsers(user, metaGroupUID){
    let member_group_list = await getMetaGroupMembersCallback(user, metaGroupUID).then(cb => cb(knex));
    let members = [];
    for(memberGroup of await member_group_list){
        members = members.concat(getGroupMemberUsers(user, metaGroupUID));
    }
};


/*
 * réflexion sur une façon possible de gérer les utilisateurs sans en demander trop à LDAP
 * Sans utilité pour le moment, ne faites pas attention 
 */
function smartUserObject(user, uid){
    this.user = user;
    this.uid = uid;

    this.resolution = {};
    let resolutionAlias = this.resolution;

    let attributeDictionnary = {
        givenName : "givenName"
    }

    for(attribute in attributeDictionnary){
        this[attribute] = function(){
            return new Promise((resolve, reject) => {
                resolutionAlias[attributeDictionnary[attribute]] = resolve;
            });
        };
    }

    this.resolve = async function(){
        let userObject = await renseignerSurUtilisateur(this.user, this.uid);
        for(attribute in this.resolution){
            this.resolution[attribute](userObject.attribute);
        }
    }
}